// Copyright (c) 2014 Intel Corporation
// All rights reserved.
//
// WARRANTY DISCLAIMER
//
// THESE MATERIALS ARE PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THESE
// MATERIALS, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Intel Corporation is the author of the Materials, and requests that all
// problem reports or change requests be submitted to it directly

#ifndef PROGRAM_FROM_RESOURCE_H
#define PROGRAM_FROM_RESOURCE_H

#include <string>
#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>
#include <Windows.h>

class ProgramFromResource : public cl::Program
{
public:
    ProgramFromResource(const std::string& name, cl_int* err = NULL)
        : Program(GetResourceString(name), err)
    {
    }

    ProgramFromResource(const std::string& name, bool build, 
                        cl_int* err = NULL)
        : Program(GetResourceString(name), build, err)
    {
    }

    ProgramFromResource(const cl::Context& context, const std::string& name, 
                        bool build = false, cl_int* err = NULL)
        : Program(context, GetResourceString(name), build, err)
    {
    }

    template <typename T>
    ProgramFromResource(const cl::Context& context, T name_begin, T name_end, 
                        bool build = false, cl_int* err = NULL)
        : Program(context, GetResourceString(name_begin, name_end), build, err)
    {
    }

protected:
    static std::string GetResourceString(const std::string& name)
    {
        // Load the OpenCL source into a string
        HMODULE module = GetModuleHandle(NULL);
        HRSRC resource = FindResourceA(module, 
                                       name.c_str(), 
                                       "OPENCL_SOURCE");
        if (!resource)
        {
            throw std::exception("Could not find resource with OpenCL "
                                 "source!");
        }
        HGLOBAL resMem = LoadResource(module, resource);
        if (!resMem)
        {
            throw std::exception("Could not load resource with OpenCL "
                                 "source!");
        }
        char* resString = (char*)LockResource(resMem);
        if (!resString)
        {
            throw std::exception("Could not lock resource with OpenCL "
                                 "source!");
        }

        DWORD resSize = SizeofResource(module, resource);
        if (!resSize)
        {
            throw std::exception("The resource with OpenCL source is "
                                 "empty!");
        }

        // MSDN says we don't have to free resource handles or the global mem 
        // handle that comes from LoadResource, so don't worry about those
        // handles.

        // Copy the source into the C++ string
        std::string source;
        return source.append(resString, resSize);
    }

    template<typename T>
    static std::string GetResourceString(T name_begin, T name_end)
    {
        // Load multiple resource strings together.
        std::string output;
        T name = name_begin;
        while (name != name_end)
        {
            output += GetResourceString(*name++);
        }
        return output;
    }
};

#endif // PROGRAM_FROM_RESOURCE_H
