/*****************************************************************************
 * Copyright (c) 2013-2015 Intel Corporation
 * All rights reserved.
 *
 * WARRANTY DISCLAIMER
 *
 * THESE MATERIALS ARE PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THESE
 * MATERIALS, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Intel Corporation is the author of the Materials, and requests that all
 * problem reports or change requests be submitted to it directly
 *****************************************************************************/
 __constant sampler_t Sampler = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_CLAMP_TO_EDGE | CLK_FILTER_NEAREST;
__kernel void Alpha( __read_only image2d_t inputImage1, __read_only image2d_t inputImage2, __write_only image2d_t outputImage, uint width, uint height, float alpha )
{
    float4 mixedImage = (0,0,0,0); 
    float4 currentPixelImage1 = (float4)(0,0,0,0); 
    float4 currentPixelImage2 = (float4)(0,0,0,0); 
    float4 calculatedPixel = (float4)(0,0,0,0); 
    float beta = 1 - alpha;

    int2 coordinates = (int2)(get_global_id(0), get_global_id(1)); 

    //read images into float4
    currentPixelImage1 = read_imagef(inputImage1, Sampler, coordinates); 
    currentPixelImage2 = read_imagef(inputImage2, Sampler, coordinates); 

    //Combine the image based on the a weighted distribution
    calculatedPixel = (currentPixelImage1 * alpha) + (currentPixelImage2 * beta); 

    //Write the image to the output
    write_imagef(outputImage, coordinates, calculatedPixel );
}
