// Copyright (c) 2015 Intel Corporation
// All rights reserved.
//
// WARRANTY DISCLAIMER
//
// THESE MATERIALS ARE PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THESE
// MATERIALS, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Intel Corporation is the author of the Materials, and requests that all
// problem reports or change requests be submitted to it directly

#define __CL_ENABLE_EXCEPTIONS
#define CL_USE_DEPRECATED_OPENCL_2_0_APIS
#define _CRT_SECURE_NO_WARNINGS 
#define _USE_MATH_DEFINES

#include <iostream>
#include <ctime>
#include <limits>
#include <cmath>
#include <math.h>
#include <CL/cl.hpp>
#include "basic.hpp"
#include "utils.h"
#include "oclobject.hpp"
#include "cmdparser.hpp"

using namespace std;

static string GetDeviceString(cl::Device device);
static bool SupportsOpenCL2(cl::Device device);
double totalTime( cl_event relEvent ); 

unsigned char *LoadImageAsBMP( const std::wstring& file_name, cl_int2 &size )
{
    const unsigned int bit32 = 32; 
    const unsigned int bit24 = 24; 
    unsigned int bytesPerPixel = 0;
    // read whole file
    std::vector<char>    input_data;
    readFile( file_name, input_data );

    // offset of the start for data
    int data_offset = *(int*)&input_data[10];
    // get image size
    size = *(cl_int2*)&input_data[18];
    // get image bits per pixel
    cl_short bits = *(cl_short*)&input_data[28];

    // calc pixel_pitch as width for 32bit images
    int pixel_pitch = size.s[0];

    if( bits != bit32 && bits != bit24 )
    {
        throw Error( "Bad format of " + wstringToString( file_name ) + " BMP file. Only 32 and 24 bits per pixel are supported" );
    }

    bytesPerPixel = bits / 8; 
    if( size.s[0] * size.s[1] * bytesPerPixel + data_offset > (int)input_data.size() )
    {
        throw Error( "Could not read " + wstringToString( file_name ) + " BMP file." );
    }
    // prepare storage for pixels
    unsigned char *p = new unsigned char[size.s[1] * pixel_pitch * bytesPerPixel];
    // convert BMP ABGR pixels into RGBA and
    // bottom-top format into top-bottom
    for( int y = 0; y<size.s[1]; ++y )
    {
        for( int x = 0; x < size.s[0]; ++x )
        {
            unsigned char* pSrc = (unsigned char*)&input_data[data_offset + ( y * size.s[0] + x ) * bytesPerPixel];
            unsigned char* pDst = p + bytesPerPixel * ( y*pixel_pitch + x );
            pDst[0] = pSrc[1];
            pDst[1] = pSrc[2];
            pDst[2] = pSrc[3];
            if( bits == 32 )
            {
                pDst[3] = pSrc[0];
            }
        }
    }
    return p;
};

void run_alpha_composting_Kernel( OpenCLBasic& oclobjects, string inputFile1, string inputFile2, string outputFile, float alpha )
{
    // Define the Variables
    bool is2Device = SupportsOpenCL2( oclobjects.device );
    cl_int          err = 0;            // opencl error code
    cl_mem          input1;             // input image2d image
    cl_mem          input2;             // input image2d image
    cl_mem          outputImage;        // output buffer that contains processed image 
    cl_image_format format;             // structure to define image format
    cl_image_format intermediate;       // structure to define image format
    cl_image_desc   desc;               // structure to define image description
    cl_int2         image_size;         // width and height of input image
    unsigned char*  input_data1 = NULL;         // input  data loaded from BMP file
    unsigned char*  input_data2 = NULL;         // input  data loaded from BMP file
    cl_ulong        start = 0, end = 0; // Variables for Time Measurement
    float           *mask = NULL;       // Array tha holds Mask Values
    double          perf_start = 0, perf_stop = 0; // Variables for Time Measurement
    double          timeFor12Composting = 0.0, timeFor20Composting = 0.0, timeForComposting = 0.0; // Variables for Time Measurement
    double          timeFor12AlphaComposting = 0.0, timeFor20AlphaComposting = 0.0; // Variables for Time Measurement

    HANDLE hConsole = GetStdHandle( STD_OUTPUT_HANDLE );
    CONSOLE_SCREEN_BUFFER_INFO consoleInfo;
    WORD saved_attributes;

    /* Save current attributes */
    GetConsoleScreenBufferInfo( hConsole, &consoleInfo );
    saved_attributes = consoleInfo.wAttributes;

    if( !is2Device )
    {
        SetConsoleTextAttribute( hConsole, FOREGROUND_RED | FOREGROUND_INTENSITY );
        printf( "\nThis Device does not Support OpenCL 2.0+. Only 1.2 Values will be printed\n" ); 
        /* Restore original attributes */
        SetConsoleTextAttribute( hConsole, saved_attributes );
    }

    //Convert String user input to safer wstring
    size_t newsize = strlen( inputFile1.c_str() ) + 1;
    size_t convertedChars = 0;
    wchar_t * wcstring1 = new wchar_t[newsize];
    mbstowcs_s( &convertedChars, wcstring1, newsize, inputFile1.c_str(), _TRUNCATE );

    newsize = strlen( inputFile2.c_str() ) + 1;
    convertedChars = 0;
    wchar_t * wcstring2 = new wchar_t[newsize];
    mbstowcs_s( &convertedChars, wcstring2, newsize, inputFile2.c_str(), _TRUNCATE );
    
    //Load the the 24/32 bit Image
    input_data1 = LoadImageAsBMP( wcstring1, image_size );
    input_data2 = LoadImageAsBMP( wcstring2, image_size );
    //STEP 1:
    SetConsoleTextAttribute( hConsole, FOREGROUND_GREEN | FOREGROUND_INTENSITY | BACKGROUND_BLUE | BACKGROUND_INTENSITY );
    printf( "\nStarting OpenCL 1.2 implementation\n" );
    //Build the opencl and build the Alpha Composting kernel
    OpenCLProgramOneKernel AlphaComposting(
        oclobjects, // opencl objects like platform, device, context and queue
        L"alphacompositing.cl",
        "",
        "Alpha");
    
    // STEP 2:
    // Create input image and output buffer
    {
        format.image_channel_data_type = CL_UNORM_INT8;
        format.image_channel_order = CL_RGBA;

        intermediate.image_channel_data_type = CL_UNORM_INT8;
        intermediate.image_channel_order = CL_RGBA;

        // init image description
        memset( &desc, 0, sizeof( desc ) ); 
        desc.image_type = CL_MEM_OBJECT_IMAGE2D; 
        desc.image_width = image_size.s[0]; 
        desc.image_height = image_size.s[1];

        // Allocate input1 image initialized by pixels loaded from .BMP File
        input1 = clCreateImage(
            oclobjects.context,
            CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR,
            &format,
            &desc,
            &input_data1[0],
            &err ); 
        SAMPLE_CHECK_ERRORS( err ); 

        // Allocate input2 image initialized by pixels loaded from .BMP File
        input2 = clCreateImage(
            oclobjects.context,
            CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
            &format,
            &desc,
            &input_data2[0],
            &err );
        SAMPLE_CHECK_ERRORS( err );

        // Allocate non-initialized output buffer 
        outputImage = clCreateImage(
            oclobjects.context,
            CL_MEM_WRITE_ONLY,
            &format,
            &desc,
            NULL,
            &err );
        SAMPLE_CHECK_ERRORS( err );
    }
    // Set the Kernel and arguments for Alpha Composting
    {
        // submit kernel to execute
        cl_event    AlphaComposting12 = NULL; // event to measure execution time

        size_t      globalsize[2] = {image_size.s[0], image_size.s[1]};

        // setup the kernel argument
        perf_start = time_stamp(); 
        err = clSetKernelArg( AlphaComposting.kernel, 0, sizeof( cl_mem ), &input1 );
        SAMPLE_CHECK_ERRORS( err );
        err = clSetKernelArg( AlphaComposting.kernel, 1, sizeof( cl_mem ), &input2 );
        SAMPLE_CHECK_ERRORS( err );
        err = clSetKernelArg( AlphaComposting.kernel, 2, sizeof( cl_mem ), &outputImage );
        SAMPLE_CHECK_ERRORS( err );
        err = clSetKernelArg( AlphaComposting.kernel, 3, sizeof( cl_uint ), &desc.image_width );
        SAMPLE_CHECK_ERRORS( err );
        err = clSetKernelArg( AlphaComposting.kernel, 4, sizeof( cl_uint ), &desc.image_height );
        SAMPLE_CHECK_ERRORS( err );
        err = clSetKernelArg( AlphaComposting.kernel, 5, sizeof( float ), &alpha );
        SAMPLE_CHECK_ERRORS( err );

        err = clEnqueueNDRangeKernel( oclobjects.queue, AlphaComposting.kernel, 2, NULL, globalsize, NULL, 0, NULL, &AlphaComposting12 );
        SAMPLE_CHECK_ERRORS( err );

        err = clWaitForEvents( 1, &AlphaComposting12 );
        SAMPLE_CHECK_ERRORS( err );
        perf_stop = time_stamp();

        timeFor12Composting = ( perf_stop - perf_start ) * 1000;

        //Profiling information: To use the profiling function in OCL, and compare the values uncomment the code below
        /*
        clGetEventProfilingInfo( AlphaComposting12, CL_PROFILING_COMMAND_START, sizeof( start ), &start, NULL );
        clGetEventProfilingInfo( AlphaComposting12, CL_PROFILING_COMMAND_END, sizeof( end ), &end, NULL );
        timeFor12AlphaComposting = (cl_double)( end - start )*(cl_double)( 1e-06 );
        */
        err = clReleaseEvent( AlphaComposting12 );
        SAMPLE_CHECK_ERRORS( err );
    }

    // STEP 4:
    // write final 1.2 result into the BMP file to be able to see result
    {
        size_t originst[3];
        size_t regionst[3];
        size_t  rowPitch = 0;
        size_t  slicePitch = 0;
        originst[0] = 0; originst[1] = 0; originst[2] = 0;
        regionst[0] = image_size.s[0]; regionst[1] = image_size.s[1]; regionst[2] = 1;

        char *p = new char[image_size.s[0]*image_size.s[1]*4];

        perf_start = time_stamp();
        err = clEnqueueReadImage(
            oclobjects.queue,
            outputImage,
            CL_TRUE,
            originst,
            regionst,
            rowPitch, 
            slicePitch,
            p,
            0,
            NULL,
            NULL );
        SAMPLE_CHECK_ERRORS( err );

        SetConsoleTextAttribute( hConsole, FOREGROUND_GREEN | FOREGROUND_INTENSITY );
        printf( "\nThe 1.2 Image will be saved as: " ); 
        printf("%s \n", outputFile.c_str() );
        SetConsoleTextAttribute( hConsole, saved_attributes );
        // save bitmap with output picture to be able to show it on monitor
        SaveImageAsBMP( (unsigned int*)p, image_size.s[0], image_size.s[1], outputFile.c_str() );

        SetConsoleTextAttribute( hConsole, FOREGROUND_GREEN | FOREGROUND_INTENSITY | BACKGROUND_BLUE | BACKGROUND_INTENSITY );
        printf( "\nOpenCL 1.2 implementation Complete\n" );
        SetConsoleTextAttribute( hConsole, saved_attributes );

        // call finish to be sure that it is finished and to be able to safely release opencl objects 
        err = clFinish( oclobjects.queue );
        SAMPLE_CHECK_ERRORS( err );
        delete[] p;
    }

    // STEP 5:
    // If the device supports OCL 2.0, Enqeue and run for it 
    {
        if( is2Device )
        {
            //Build the opencl and build the Alpha Composting kernel
            OpenCLProgramOneKernel AlphaComposting20(
                oclobjects, // opencl objects like platform, device, context and queue
                L"alphacompositing20.cl",
                "",
                "Alpha",
                "-cl-std=CL2.0" );

            cl_event    AlphaComposting20Execution = NULL; // event to measure execution time
            size_t      globalsize[2] = {image_size.s[0], image_size.s[1]};

            //setup the kernel argument
            perf_start = time_stamp();
            err = clSetKernelArg( AlphaComposting20.kernel, 0, sizeof( cl_mem ), &input1 );
            SAMPLE_CHECK_ERRORS( err );
            err = clSetKernelArg( AlphaComposting20.kernel, 1, sizeof( cl_mem ), &input2 );
            SAMPLE_CHECK_ERRORS( err );
            err = clSetKernelArg( AlphaComposting20.kernel, 2, sizeof( cl_uint ), &desc.image_width );
            SAMPLE_CHECK_ERRORS( err );
            err = clSetKernelArg( AlphaComposting20.kernel, 3, sizeof( cl_uint ), &desc.image_height );
            SAMPLE_CHECK_ERRORS( err );
            err = clSetKernelArg( AlphaComposting20.kernel, 4, sizeof( float ), &alpha );
            SAMPLE_CHECK_ERRORS( err );

            // submit kernel to execute
            err = clEnqueueNDRangeKernel( oclobjects.queue, AlphaComposting20.kernel, 2, NULL, globalsize, NULL, 0, NULL, &AlphaComposting20Execution );
            SAMPLE_CHECK_ERRORS( err );
            err = clWaitForEvents( 1, &AlphaComposting20Execution );
            SAMPLE_CHECK_ERRORS( err );
            
            perf_stop = time_stamp();

            timeFor20Composting = ( perf_stop - perf_start ) * 1000;

            //Profiling information: To use the profiling function in OCL, and compare the values uncomment the code below
            /*
            clGetEventProfilingInfo( AlphaComposting20Execution, CL_PROFILING_COMMAND_START, sizeof( start ), &start, NULL );
            clGetEventProfilingInfo( AlphaComposting20Execution, CL_PROFILING_COMMAND_END, sizeof( end ), &end, NULL );
            timeFor20AlphaComposting = (cl_double)( end - start )*(cl_double)( 1e-06 );
            */
            err = clReleaseEvent( AlphaComposting20Execution );
            SAMPLE_CHECK_ERRORS( err );
        }
    }
    
    // STEP 7:
    // write final 2.0 result into the BMP file to be able to see result
    {
        if( is2Device )
        {
            SetConsoleTextAttribute( hConsole, FOREGROUND_GREEN | FOREGROUND_INTENSITY | BACKGROUND_BLUE | BACKGROUND_INTENSITY );
            printf( "\nStarting OpenCL 2.0 implementation\n" );
            size_t originst[3];
            size_t regionst[3];
            size_t  rowPitch = 0;
            size_t  slicePitch = 0;
            originst[0] = 0; originst[1] = 0; originst[2] = 0;
            regionst[0] = image_size.s[0]; regionst[1] = image_size.s[1]; regionst[2] = 1;

            char *p = new char[image_size.s[0] * image_size.s[1] * 4];

            err = clEnqueueReadImage(
                oclobjects.queue,
                input1,
                CL_TRUE,
                originst,
                regionst,
                rowPitch,
                slicePitch,
                p,
                0,
                NULL,
                NULL );
            SAMPLE_CHECK_ERRORS( err );
            string outputfile20 = "images\\20_outputFile.bmp";

            SetConsoleTextAttribute( hConsole, FOREGROUND_GREEN | FOREGROUND_INTENSITY );
            printf( "\nThe 2.0 Image will be saved as: " );
            printf( "%s \n", outputfile20.c_str() );
            SetConsoleTextAttribute( hConsole, saved_attributes );

            // save bitmap with output picture to be able to show it on monitor
            SaveImageAsBMP( (unsigned int*)p, image_size.s[0], image_size.s[1], outputfile20.c_str() );

            SetConsoleTextAttribute( hConsole, FOREGROUND_GREEN | FOREGROUND_INTENSITY | BACKGROUND_BLUE | BACKGROUND_INTENSITY );
            printf( "\nOpenCL 2.0 implementation Complete\n" );

            SetConsoleTextAttribute( hConsole, saved_attributes );
            // call finish to be sure that unmap is finished and to be able to safely release opencl objects 
            err = clFinish( oclobjects.queue );
            SAMPLE_CHECK_ERRORS( err );
            delete[] p;
        }
    }

    //Release the various Buffers
    delete[] input_data1; 
    delete[] input_data2;
    delete[] mask; 
    clReleaseMemObject( outputImage );
    clReleaseMemObject( input1 );
    clReleaseMemObject( input2 );
}

int main(int argc, const char** argv)
{
    string inputFile1 = ""; 
    string inputFile2 = "";
    string outputFile = ""; 
    float alpha = 0.0; 
    string temp = ""; 
    int gamma = 0; 
    try
    {
        // Define and parse command-line arguments.
        CmdParserCommon cmd( argc, argv );
        cmd.parse();
        // Immediatly exit if user wanted to see the usage information only.
        if( cmd.help.isSet() )
        {
            return EXIT_SUCCESS;
        }

        // Create the necessary OpenCL objects up to device queue.
        // we use CL_QUEUE_PROFILING_ENABLE to be able geting kernel execution time to comapre sRGBA and RGBA performance
        OpenCLBasic oclobjects(
            cmd.platform.getValue(),
            cmd.device_type.getValue(),
            cmd.device.getValue(),
            CL_QUEUE_PROFILING_ENABLE );

        temp = cmd.alpha.getValue(); 
        alpha = (float)atof( temp.c_str() );
        if( alpha < 0.0 || alpha > 1.0 ){
            printf( "\nFAILURE: Value of alpha must be between 0 and 1. \nValue of alpha %.6f", alpha );
            return EXIT_FAILURE;
        }

        inputFile1 = cmd.input1_Filename.getValue(); 
        if( inputFile1.substr( inputFile1.find_last_of( "." ) + 1 ) != "bmp" )
        {
            printf( "\nFAILURE: Expected input file format is a 32-bit bmp file" ); 
            return EXIT_FAILURE; 
        }

        inputFile2 = cmd.input2_Filename.getValue();
        if( inputFile2.substr( inputFile2.find_last_of( "." ) + 1 ) != "bmp" )
        {
            printf( "\nFAILURE: Expected input file format is a 32-bit bmp file" );
            return EXIT_FAILURE;
        }

        outputFile = cmd.output_Filename.getValue();
        if( outputFile.substr( outputFile.find_last_of( "." ) + 1 ) != "bmp" )
        {
            printf( "\nWARNING: Expected output file format is a 32-bit bmp file." );
        }

        // run picture processing using RGBA image format and print execution time
        printf( "\nRun Alpha Composition on input Image...\n" );
        run_alpha_composting_Kernel( oclobjects, inputFile1, inputFile2, outputFile, alpha );
        return EXIT_SUCCESS;
    }
    catch( const Error& error )
    {
        cerr << "[ ERROR ] Sample application specific error: " << error.what() << "\n";
        return EXIT_FAILURE;
    }
    catch( const exception& error )
    {
        cerr << "[ ERROR ] " << error.what() << "\n";
        return EXIT_FAILURE;
    }
    catch( ... )
    {
        cerr << "[ ERROR ] Unknown/internal error happened.\n";
        return EXIT_FAILURE;
    }
}

string GetDeviceString(cl::Device device)
{
    string device_string;
    device_string += string("Device: ") + 
                     device.getInfo<CL_DEVICE_NAME>() + 
                     string("\nVendor: ") +
                     device.getInfo<CL_DEVICE_VENDOR>();
    return device_string;
}

bool SupportsOpenCL2(cl::Device device)
{
    // Checks if the device supports OpenCL 2.0
    string dev_version = device.getInfo<CL_DEVICE_VERSION>();
    // The format of the version string is defined in the spec as
    // "OpenCL <major>.<minor> <vendor-specific>"
    dev_version = dev_version.substr(string("OpenCL ").length());
    dev_version = dev_version.substr(0, dev_version.find('.'));

    string lang_version = device.getInfo<CL_DEVICE_OPENCL_C_VERSION>();
    // The format of the version string is defined in the spec as
    // "OpenCL C <major>.<minor> <vendor-specific>"
    lang_version = lang_version.substr(string("OpenCL C ").length());
    lang_version = lang_version.substr(0, lang_version.find('.'));

    return (stoi(dev_version) >= 2 && stoi(lang_version) >= 2);
}

cl::Device GetDevice()
{
    // Find the GPU device if there is one.  Otherwise, use the CPU.
    cl::Platform platform = platform.getDefault();
    vector<cl::Device> devices;
    try
    {
        platform.getDevices(CL_DEVICE_TYPE_GPU, &devices);
    }
    catch (cl::Error)
    {
        platform.getDevices(CL_DEVICE_TYPE_CPU, &devices);
    }
    return devices[0];
}

